module 'data.atom' {
   constructor: '=>'/2.

   /** check(X).
    *  Succeeds if <code>X</code> is instantiated to an atom.
    */

   check(Atom) :-
      'kernel':atom_check(Atom).

   /** length(Atom, Length).
    *  <prerequisite>
    *  <code>Atom</code> is an instantiated atom.
    *  </prerequisite>
    *  Succeeds if the length of <code>Atom</code> is <code>Length</code>.
    */

   length(Atom, Length) :-
      'kernel':atom_length(Atom, Length).

   /** concat(A, B, R).
    *  <prerequisite>
    *  At least two of <code>A</code>, <code>B</code> and
    *  <code>R</code> are instantiated atoms.
    *  </prerequisite>
    *  Succeeds if <code>R</code> is the concatenation of <code>A</code>
    *  and <code>B</code>.
    */

   concat(A, B, R) :-
      'kernel':atom_concat(A, B, R).

   /** new(Atom).
    *  <prerequisite>
    *  <code>Atom</code> is uninstantiated.
    *  </prerequisite>
    *  <code>Atom</code> is unified with a fresh atom.
    */

   new(Atom) :-
      'kernel':atom_new(Atom).

   /** new_backtrackable(Atom).
    *  <prerequisite>
    *  <code>Atom</code> is uninstantiated.
    *  </prerequisite>
    *  <code>Atom</code> is unified with a fresh atom.  Generated atoms are
    *  reused on backtrack.
    */

   new_backtrackable(Atom) :-
      'kernel':atom_new_backtrackable(Atom).

   /** string(Atom, String).
    *  <prerequisite>
    *  <code>Atom</code> is instantiated to an atom.
    *  </prerequisite>
    *  Succeeds if <code>String</code> is the string containing to
    *  the name of <code>Atom</code>.
    */

   string(Atom) = kernel:atom_to_string(Atom).

   stream(Atom) = data.list:stream(string(Atom)).

   /** sub(Atom, Before, Length, After, SubAtom).
    *  <prerequisite>
    *  <code>Atom</code> is an instantiated atom.
    *  </prerequisite>
    *  <prerequisite>
    *  Either <code>SubAtom</code> is an instantiated atom or at least
    *  two of <code>Before</code>, <code>Length</code> and
    *  <code>After</code> are instantiated integers.
    *  </prerequisite>
    *  Succeeds if <code>Atom</code> is the concatenation of
    *  atoms <code>Left</code>, <code>SubAtom</code> and
    *  <code>Right</code>, with respective lengths
    *  <code>Before</code>, <code>Length</code> and
    *  <code>After</code>.
    */

   sub(Atom, Before, Length, After, SubAtom) :-
      kernel:atom_sub(Atom, Before, Length, After, SubAtom).

   suffix(Atom, Length, Suffix) :-
      sub(Atom, _, Length, 0, Suffix).

   remove_suffix(Atom, Suffix, SubAtom) :-
      (
         concat(SubAtom, Suffix, Atom)
      ->
         true
      ;
         Atom = SubAtom
      ).

   /** format(Format, Args, Atom).
    *  <prerequisite>
    *  <code>Format</code> is a format specification and
    *  <code>Args</code> is a suitable list of arguments.
    *  </prerequisite>
    *  Succeeds if <code>Atom</code> contains the formatted
    *  text.
    */

   format(Format, Args, Atom) :-
      'kernel':atom_format(Format, Args, Atom).

   /** join(Atoms, Delim, Atom).
    *  <prerequisite>
    *  <code>Atoms</code> is a list of instantiated atoms.
    *  <code>Delim</code> is an instantiated atom.
    *  </prerequisite>
    *  Succeeds if <code>Atom</code> is the concatenation
    *  of atoms enumerated in <code>Atoms</code>, separated
    *  with <code>Delim</code>.
    */

   join(Atoms, Delim, Atom) :-
      'data.atom':string(Delim, DelimStr),
      module Convert [] {
         do(Atom, String) :-
            'data.atom':string(Atom, String).
      },
      'data.list':join('data.list':iter(Atoms, Convert), DelimStr, String),
      'data.string':atom(String, Atom).

   /** split(Atom, Delim, Atoms).
    *  <prerequisite>
    *  <code>Atoms</code> is a list of instantiated atoms.
    *  </prerequisite>
    *  <prerequisite>
    *  <code>Delim</code> is an instantiated character.
    *  </prerequisite>
    *  Succeeds if <code>Atoms</code> is the list of
    *  sub-atoms of <code>Atom</code> delimited
    *  with <code>Delim</code>.
    */

   split(Atom, Delim, Atoms) :-
      split_from(0, 0, Atom, Delim, Atoms).

   split_from(Origin, Index, Atom, Delim, Atoms) :-
      length(Atom, Length),
      (
         'data.number':(Index < Length)
      ->
         'data.number':succ(Index, SuccIndex),
         (
            sub(Atom, Index, 1, _, Delim)
         ->
            sub(Atom, Origin, 'data.number':(Index - Origin), _, C),
            Atoms = [C | Tail],
            split_from(SuccIndex, SuccIndex, Atom, Delim, Tail)
         ;
            split_from(Origin, SuccIndex, Atom, Delim, Atoms)
         )
      ;
         sub(Atom, Origin, _, 0, C),
         Atoms = [C]
      ).

   /** split_at_delim_left(Atom, Delim, Left, Right).
    *  <prerequisite>
    *  <code>Atom</code> is an instantiated atom.
    *  </prerequisite>
    *  <prerequisite>
    *  <code>Delim</code> is an instantiated character.
    *  </prerequisite>
    *  Succeeds if <code>Atom</code> contains at least
    *  one occurrence of <code>Delim</code>, and
    *  <code>Left</code> is the prefix before the left-most
    *  occurrence, and <code>Right</code> the suffix
    *  starting immediately after it.
    */

   split_at_delim_left(Atom, Delim, Left, Right) :-
      split_at_delim_left_from(0, Atom, Delim, Left, Right).

   split_at_delim_left_from(Index, Atom, Delim, Left, Right) :-
      length(Atom, Length),
      (
         'data.number':(Index < Length)
      ->
         'data.number':succ(Index, SuccIndex),
         (
            sub(Atom, Index, 1, _, Delim)
         ->
            sub(Atom, 0, Index, _, Left),
            sub(Atom, SuccIndex, _, 0, Right)
         ;
            split_at_delim_left_from(
               SuccIndex,
               Atom,
               Delim,
               Left,
               Right
            )
         )
      ;
         fail
      ).

   /** split_at_delim_right(Atom, Delim, Left, Right).
    *  <prerequisite>
    *  <code>Atom</code> is an instantiated atom.
    *  </prerequisite>
    *  <prerequisite>
    *  <code>Delim</code> is an instantiated character.
    *  </prerequisite>
    *  Succeeds if <code>Atom</code> contains at least
    *  one occurrence of <code>Delim</code>, and
    *  <code>Left</code> is the prefix before the right-most
    *  occurrence, and <code>Right</code> the suffix
    *  starting immediately after it.
    */

   split_at_delim_right(Atom, Delim, Left, Right) :-
      length(Atom, Length),
      split_at_delim_right_from(
         'data.number':pred(Length), Atom, Delim, Left, Right
      ).

   split_at_delim_right_from(Index, Atom, Delim, Left, Right) :-
      (
         'data.number':(Index >= 0)
      ->
         (
            sub(Atom, Index, 1, _, Delim)
         ->
            sub(Atom, 0, Index, _, Left),
            sub(Atom, 'data.number':succ(Index), _, 0, Right)
         ;
            split_at_delim_right_from(
               'data.number':pred(Index),
               Atom,
               Delim,
               Left,
               Right
            )
         )
      ;
         fail
      ).

   /** subst(Src, Table, Tgt).
       <prerequisite>
       <code>Src</code> is an atom.
       </prerequisite>
       <prerequisite>
       <code>Table</code> is a list with elements of the form
       <code>A => B</code> where <code>A</code> and <code>B</code>
       are atoms.
       </prerequisite>
       Succeeds if <code>Tgt</code> is the atom obtained from
       <code>Src</code> by substituting every occurrences of
       sub-atom <code>A</code> in <code>Src</code> by <code>B</code>,
       for every <code>A => B</code> in <code>Table</code>.
    */

   syntax: infix '=>'.

   syntax: '=>' < ':'.

   subst(Src, Table, Tgt) :-
      Tgt =
         'data.string':atom(
            'data.list':subst(
               'data.atom':string(Src),
               'data.list':iter(Table) [] {
                  do(From => To) =
                     ('data.atom':string(From) => 'data.atom':string(To)).
               }
            )
         ).

   number(Atom) = 'kernel':atom_to_number(Atom).

   lower_upper(Lower, Upper) :-
      kernel:char_lower_upper(Lower, Upper).

   capitalize(Atom) =
      (
         Atom = ''
      ->
         ''
      ;
         (Initial = sub(Atom, 0, 1, _)) >>
         (Suffix = sub(Atom, 1, _, 0)) >>
         (Initial_upper = lower_upper(Initial)) >>
         concat(Initial_upper, Suffix)
      ).
}
